#include "pqrst.h"
#include "unittests.h"

#include <stdio.h>
#include <stdlib.h>

#ifndef DEBUG
#define DEBUG 0
#endif

// Note: we use the override keyword below
// -- we are assuming that this code is being built by at least a C++11 compiler.


// half-way through the time range
#define HALF_TIME 0x80000000
// the largest time value
#define FULL_TIME 0xffffffff

// suppress unused-parameter warnings
// (this works for both clang and GCC, but GCC demands the __attribute__ come after the variable)
#define UNUSED(x) UNUSED_ ## x __attribute__((__unused__))



// We want this program to test both the self-queue and global-queue
// variants, and this involves a certain amount of macro gymnastics.
#if PQRST_SELF_QUEUE
// A queue...
TaskQueue test_queue;
#define Q test_queue

// declaration
#define TQP(tqp) TaskQueue* tqp,
// reference to argument
#define TQPX(tqp) tqp,
// ...as sole argument
#define TQP1(tqp) tqp
// pointer to queue, as an argument
#define TQPP &Q,
#else
// the queue called Queue is declared within pqrst.cpp
#define Q Queue

#define TQP(tqp)
#define TQPX(tqp)
#define TQP1(tqp)
#define TQPP
#endif


#define BUFFER_LENGTH 128
char log_buffer[BUFFER_LENGTH];
char* logbufp = log_buffer;
void reset_all(void)
{
    logbufp = log_buffer;
    *logbufp = '\0';
    Q.remove_all_tasks();
    Q.initialise();
}
void buf_append_ident(const char* s, ms_t t)
{
    if (logbufp != log_buffer) {
        *logbufp++ = ',';
    }

    size_t avail = &log_buffer[BUFFER_LENGTH] - logbufp;
    size_t nw;
    if (t >= 0x10000) {
        // for high times, it's more useful to have the times shown in hex
        nw = snprintf(logbufp, avail, "%s@0x%x", s, t);
    } else {
        nw = snprintf(logbufp, avail, "%s@" MS_FMT, s, t);
    }

    if (nw >= avail) {
        fputs("log buffer too short!\n", stderr);
        exit(1);
    }

    logbufp += nw;
}
void assert_log(const char* expected)
{
    assert_equal(expected, log_buffer);
}

class TestTask : public Task {
private:
    const char* ident_;
public:
    TestTask(TQP(tqp) const char* ident) : Task(TQP1(tqp)), ident_(ident) {};
    void run(ms_t t) override { buf_append_ident(ident_, t); };
    const char* ident(void) const { return ident_; };
    static int times2(int i) { return 2*i; }; // times2 is a static method
};

// TestLoopTask will repeat itself 'n' times separated by 'cadence'
class TestLoopTask : public LoopTask {
private:
    const char* ident_;

public:
    TestLoopTask(TQP(tqp) const char* ident, ms_t cadence, int n=-1)
        : LoopTask(TQPX(tqp) cadence, n),
          ident_(ident) {};

    void run(ms_t t) override {
        buf_append_ident(ident_, t);
#if DEBUG
        if (t >= 0x10000) {
            fprintf(stderr, "TestLoopTask %s at 0x%x\n", ident_, t);
        } else {
            fprintf(stderr, "TestLoopTask %s at " MS_FMT "\n", ident_, t);
        }
#endif
    };

    const char* ident(void) const { return ident_; };
};

// Helper task.  When this task runs, it removes all of the tasks from
// the queue.  It has a no-arg constructor in both the
// PQRST_SELF_QUEUE cases.
//
// Usage:
//    TerminateTask x;
//    x.start(20);  // clear the queue at time 20
class TerminateTask : public Task {
public:
#if PQRST_SELF_QUEUE
    TerminateTask(void) : Task(&Q) {};
#else
    TerminateTask(void) : Task() {};
#endif
    void run(ms_t UNUSED(t)) { Q.remove_all_tasks(); };
};

// Run the queue until it's empty, incrementing the time between each Q.run_ready().
static void run_queue(int init=0, int increment=1)
{
    for (ms_t t=init; !Q.is_empty(); t+=increment) {
        Q.run_ready(t);
    }
}

// As with run_queue, but only incrementing the time if nothing ran.
// This means tasks scheduled for the same time will run at the same
// time, and is useful to avoid confusion/complication when running
// over the half-way maintenance task
static void run_queue_with_repeats(int init=0, int increment=1)
{
    for (ms_t t=init; !Q.is_empty(); ) {
        if (! Q.run_ready(t)) t+=increment;
    }
}

static void run_queue_to(ms_t from, ms_t to, int increment=1)
{
    for (ms_t t=from; t<=to; t += increment) {
        Q.run_ready(t);
    }
}

void test_in_order(void)
{
    reset_all();

    assert_equal(0, Q.size());

    TestTask t1 = TestTask(TQPP "t1");
    TestTask t2 = TestTask(TQPP "t2");
    TestTask t3 = TestTask(TQPP "t3");
    TestTask t4 = TestTask(TQPP "t4");
    t1.start(2);
    t2.start(4);
    t3.start(6);
    t4.start(8);

    assert_equal(4, Q.size());

    run_queue();

    assert_log("t1@2,t2@4,t3@6,t4@8");
}

void test_function_task(void)
{
    reset_all();

    TestTask t1 = TestTask(TQPP "t1");

    // Task f1 does not repeat
    FunctionTask(TQPP [](ms_t t) { buf_append_ident("f1", t); }).start(6);
    // Task f2 runs twice
    FunctionTask(TQPP [](ms_t t) { buf_append_ident("f2", t); }, 5, 2).run_after(4);

    t1.start(2);

    run_queue();

    assert_log("t1@2,f2@4,f1@6,f2@9");
}

TestTask global1 = TestTask(TQPP "g1");
void test_function_task2(void)
{
    // test referring to a global object (doesn't need a closure)
    // and calling a class method.
    reset_all();

    FunctionTask f1 = FunctionTask(TQPP [](ms_t t) {
            buf_append_ident("f1", t);
            global1.run_at(7);
        });
    FunctionTask f2 = FunctionTask(TQPP [](ms_t t) {
            buf_append_ident("f2", t);
            buf_append_ident("x2", TestTask::times2(t));
        }, 5, 2);

    f1.start(3);
    f2.start();

    run_queue();

    assert_log("f2@0,x2@0,f1@3,f2@5,x2@10,g1@7");
}

void test_out_of_order(void)
{
    reset_all();

    TestTask t1 = TestTask(TQPP "t1");
    TestTask t2 = TestTask(TQPP "t2");
    TestTask t3 = TestTask(TQPP "t3");
    TestTask t4 = TestTask(TQPP "t4");

    t1.start(4);
    t2.start(2);              // inserted first
    t3.start(16);             // inserted last
    t4.start(6);              // inserted in the middle


    for (ms_t t = 0; !Q.is_empty(); t++) {
        Q.run_ready(t);

        switch (t) {
          case 7:
            t1.run_immediately(); // due immediately
            break;

          case 9:
            t1.run_at(8);       // due earlier than now
            break;

          case 13:
            t2.run_after(0);    // due immediately
            break;
        }
    }

    assert_log("t2@2,t1@4,t4@6,t1@8,t1@10,t2@14,t3@16");
}

void test_removals(void)
{
    reset_all();

    TestTask t1 = TestTask(TQPP "t1");
    TestTask t2 = TestTask(TQPP "t2");
    TestTask t3 = TestTask(TQPP "t3");
    TestTask t4 = TestTask(TQPP "t4");
    TestTask t5 = TestTask(TQPP "t5");

    t1.cancel();                // cancelled while queue is empty

    t1.start(2);
    t2.start(4);
    t3.start(6);
    t4.start(8);

    t1.cancel();                // head of queue
    t3.cancel();                // middle of queue
    t4.cancel();                // end of queue

    t5.cancel();                // unqueued, and cancelled when queue is non-empty

    run_queue();

    assert_log("t2@4");
}

void test_uncancel(void)
{
    reset_all();

    TestTask t1 = TestTask(TQPP "t1");
    TestTask t2 = TestTask(TQPP "t2");

    t1.start(5);
    t2.start(15);

    t2.uncancel();              // should be a no-op

    assert_equal(5, t1.cancel());
    assert_equal(15, t2.cancel());

    run_queue_to(0, 7);

    // assert_equal(5, t1.uncancel());
    t1.uncancel();
    assert_equal(15, t2.uncancel());

    run_queue(8);

    assert_log("t1@8,t2@15");
}

void test_remove_all(void)
{
    reset_all();

    TestTask t1 = TestTask(TQPP "t1");
    TestTask t2 = TestTask(TQPP "t2");
    TestTask t3 = TestTask(TQPP "t3");
    TestTask t4 = TestTask(TQPP "t4");

    t1.start(2);
    t2.start(4);
    t3.start(6);
    t4.start(8);

    for (ms_t t = 0; !Q.is_empty(); t++) {
        Q.run_ready(t);

        if (t == 5) {
            Q.remove_all_tasks();
        }
    }

    assert_log("t1@2,t2@4");
}

void test_weird_calls(void)
{
    reset_all();

    Q.push_at(nullptr, 1);      // shouldn't segfault
    Q.push_after(nullptr, 1);   // ditto
    Q.remove(nullptr); // no-op
    assert_pass();
}

void test_ordering(void)
{
    // if a task is pushed on to the queue with the same due time as
    // an earlier one, it is the earlier one which should be removed first

    reset_all();
    TestTask t1 = TestTask(TQPP "t1");
    TestTask t2 = TestTask(TQPP "t2");

    t1.start(2);
    t2.start(2);
    run_queue();
    assert_log("t1@2,t2@3");

    reset_all();
    t2.start(2);
    t1.start(2);
    run_queue();
    assert_log("t2@2,t1@3");
}

void test_looping(void)
{
    reset_all();

    TestTask t1 = TestTask(TQPP "t1");
    TestTask t2 = TestTask(TQPP "t2");
    // Task rt will run only three times
    TestLoopTask rt = TestLoopTask(TQPP "r1", 4, 3);

    rt.start(3);

    t1.start(5);                // interpolate Tasks t1 and t2
    t2.start(10);

    run_queue();

    assert_log("r1@3,t1@5,r1@7,t2@10,r1@11");
}

void test_indefinite_loop(void)
{
    reset_all();

    // Task rt will run indefinitely unless stopped
    TestLoopTask rt = TestLoopTask(TQPP "r1", 5);
    rt.start();

    for (ms_t t = 0; !Q.is_empty(); t++) {
        switch (t) {
          case 16:
            assert_equal(rt.set_cadence(3), 5); // reset cadence to 3, from 5
            break;

          case 28:
            rt.cancel();
            break;
        }

        Q.run_ready(t);
    }

    assert_log("r1@0,r1@5,r1@10,r1@15,r1@20,r1@23,r1@26");
}

// Test semantics of TaskQueue::push_at.
// The run method in LoopTask calls push_after, which refers to
// the reftime, which is set, when the task is dequeued, to the
// scheduled ready_time.  That ready_time is set when the Task is
// first enqueued, within push_at (at time 8 below).
void test_reference_times(void)
{
    reset_all();

    // we add a task which runs just before the time 8 in the loop,
    // so that the queue's reftime is set
    TestTask t1 = TestTask(TQPP "t1");
    t1.start(8);

    TestLoopTask r1 = TestLoopTask(TQPP "r1", 10);
    TestLoopTask r2 = TestLoopTask(TQPP "r2", 10);
    TestLoopTask r3 = TestLoopTask(TQPP "r3", 10);
    TestLoopTask r4 = TestLoopTask(TQPP "r4", 10);

    TerminateTask x;
    x.start(30);

    for (ms_t t=0; !Q.is_empty(); t++) {
        Q.run_ready(t);

        if (t == 8) {
            // r1 has a due time of 4 (so repeats at 4+10, ...)
            Q.push_at(&r1, 4);
            // r2 has due time of 8 (so repeats at 8+10, ...)
            Q.push_after(&r2, 0);
            // r3 also has a due time of 8 (so repeats at 8+10, ...),
            // but push_immediately() forces it to the head of the
            // queue, ahead of r1 and r2.
            //
            // Therefore r3 runs at time 9, r1 at 10 and r2 at 11.
            Q.push_immediately(&r3);
            // r4 runs at time 12, so doesn't participate in any of
            // this squabbling
            Q.push_at(&r4, 12);
        }
    }

    assert_log("t1@8,r3@9,r1@10,r2@11,r4@12,r1@14,r3@18,r2@19,r4@22,r1@24,r3@28,r2@29");
}

void test_reference_times_slippy(void)
{
    reset_all();

    TestLoopTask r1 = TestLoopTask(TQPP "r1", 10);
    TestLoopTask r2 = TestLoopTask(TQPP "r2", 10);
    r2.set_task_slippy(true);

    r1.start();
    r2.start(5);

    ms_t run_times[] = {
        0, 1, 2, 3, 4, 5, 6, 7,
            // both tasks will now have run, and r1 should be
            // rescheduled for t=10, and r2 for t=15.

            12,
            // r1 is late being rescheduled, but should be scheduled
            // for due-time+10 = 20.

            18,
            // r2 is late being rescheduled, but should be scheduled
            // for 'now'+10 = 28

            19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29,
            };
    int n_times = sizeof(run_times)/sizeof(run_times[0]);

    for (int i=0; i<n_times; i++) {
        Q.run_ready(run_times[i]);
    }

    assert_log("r1@0,r2@5,r1@12,r2@18,r1@20,r2@28");
}

void test_loop_stutter(void)
{
    class StutteringLoopTask : public TestLoopTask {
        // A LoopTask which 'stutters' -- rescheduling itself early -- at
        // a specified time.
    private:
        ms_t stutter_time;
        bool have_stuttered;
    public:
        StutteringLoopTask(TQP(tqp) const char* ident, ms_t cadence, ms_t stutter_time)
            : TestLoopTask(TQPX(tqp) ident, cadence),
              stutter_time(stutter_time),
              have_stuttered(false) {};

        void run_task(ms_t t)
        {
            if (! have_stuttered && t >= stutter_time) {
                ms_t orig_cadence = set_cadence(2);
                LoopTask::run_task(t);
                set_cadence(orig_cadence);
                have_stuttered = true;
            } else {
                LoopTask::run_task(t);
            }
        };
    };

    // Test that we can have a loop task re-run itself early.
    // This also acts as a test of overriding run_task().
    reset_all();

    StutteringLoopTask r1 = StutteringLoopTask(TQPP "r1", 10, 30);
    r1.start();

    TerminateTask().start(60);

    run_queue();

    assert_log("r1@0,r1@10,r1@20,r1@30,r1@32,r1@42,r1@52");
}

// check that one task can cancel another

class CancelTask : public Task {
    // a task to cancel another task at a given time
private:
    LoopTask* other_;           // the task to cancel
public:
    CancelTask(TQP(tqp) LoopTask* o);
    void run(ms_t) override;
};

CancelTask::CancelTask(TQP(tqp) LoopTask* o)
    : Task(TQP1(tqp)),
      other_(o)
{
    // nothing
}

void CancelTask::run(ms_t UNUSED(t))
{
    other_->cancel();
}

void test_indefinite_loop_pair(void)
{
    reset_all();

    TestLoopTask rt = TestLoopTask(TQPP "r1", 4);
    CancelTask ct = CancelTask(TQPP &rt);

    rt.start();
    ct.start(20);

    run_queue();

    assert_log("r1@0,r1@4,r1@8,r1@12,r1@16");
}

// check that a task can cancel itself by calling CancelTask

void test_self_cancel_task(void)
{
    class SelfCancelTask : public LoopTask {
    private:
        CancelTask ct_;
        ms_t endtime_;
        const char* ident_;
    public:
        SelfCancelTask(TQP(tqp) const char* ident, ms_t cadence, ms_t endtime)
            : LoopTask(TQPX(tqp) cadence),
              ct_(TQPX(tqp) this),
              endtime_(endtime),
              ident_(ident) {};

        void start(void) override {
            LoopTask::start();
            run_at(endtime_, &ct_);
        };

        void run(ms_t t) override {
            buf_append_ident(ident_, t);
        };
    };

    reset_all();

    SelfCancelTask sct(TQPP "sct", 4, 9);
    sct.start();

    run_queue();

    assert_log("sct@0,sct@4,sct@8");
}

void test_self_cancel_task2(void)
{
    // ...and by calling cancel on itself
    class SelfCancelTask : public LoopTask {
    private:
        ms_t endtime_;
        const char* ident_;
    public:
        SelfCancelTask(TQP(tqp) const char* ident, ms_t cadence, ms_t endtime)
            : LoopTask(TQPX(tqp) cadence),
              endtime_(endtime),
              ident_(ident) {};

        void run(ms_t t) override {
            buf_append_ident(ident_, t);
            if (t > endtime_) {
                cancel();
            }
        };
    };

    reset_all();

    SelfCancelTask sct(TQPP "sct", 4, 9);
    sct.start();

    TerminateTask x;            // just in case...
    x.start(20);

    run_queue();

    assert_log("sct@0,sct@4,sct@8,sct@12");
}


// check we can change the repeat-count mid-loop
void test_adjusted_loop(void)
{
    reset_all();

    TestLoopTask rt = TestLoopTask(TQPP "r1", 4);
    // we are initial set to loop indefinitely
    rt.start();

    for (ms_t t = 0; !Q.is_empty(); t++) {
        if (t == 18) {
            // do only 2 more repeats
            assert_equal(rt.set_repeats(2), -1);
        }

        Q.run_ready(t);
    }

    assert_equal("r1@0,r1@4,r1@8,r1@12,r1@16," // repeats set to 2 here
                 "r1@20,r1@24",
                 log_buffer);
}

// check the behaviour of a LoopTask with a zero repeat count -- it
// should be identical to a Task

void test_nonloop_loop(void)
{
    // This UnloopingLoopTask repeats itself 'by hand', decrementing a
    // counter, n, and rescheduling itself after a time 4*n.  Thus we
    // should end up with a sequence of runs at intervals
    // (..., 16, 9, 4, 1)*4, and no others.
    class UnloopingLoopTask : public LoopTask {
    private:
        const char* ident_;
        int n;
    public:
        UnloopingLoopTask(TQP(tqp) const char* ident, int my_repeat_count)
            : LoopTask(TQPX(tqp) 10, 0), // a loop-task with a repeat count of zero
              ident_(ident),
              n(my_repeat_count) {};

        void start(void) override { LoopTask::start(); };

        void run(ms_t t) override {
            buf_append_ident(ident_, t);

            if (n > 0) {
                run_after(4 * n, this);
                n--;
            }
        };
    };

    reset_all();

    UnloopingLoopTask T = UnloopingLoopTask(TQPP "t", 3);
    T.start();

    run_queue();

    assert_log("t@0,t@12,t@20,t@24");
}

void test_zerocadence_loop(void)
{
    reset_all();

    TestLoopTask t1 = TestLoopTask(TQPP "t1", 4, -1);
    t1.start();

    for (ms_t t=0; t<20; t++) {
        if (t == 10) {
            // set the cadence to zero -- this means that the loop
            // will not be requeued after the next run
            t1.set_cadence(0);
        }

        Q.run_ready(t);
    }

    assert_log("t1@0,t1@4,t1@8,t1@12");
}

void test_signal_when_empty(void)
{
    reset_all();                // queue is now empty
    Q.signal(1, 0);             // this shouldn't cause a segfault
    assert_pass();              // well done, us!
}

class Signaller {
private:
    char signal_ident_[10];
public:
    Signaller(const char* ident);
    void signal(int, ms_t);
};
Signaller::Signaller(const char* ident)
{
    snprintf(signal_ident_, 10, "S%s", ident);
}
void Signaller::signal(int UNUSED(signal), ms_t t)
{
    //fprintf(stderr, "Signaller::signal @ " MS_FMT "\n", t);
    buf_append_ident(signal_ident_, t);
}

class SignalTask : public TestLoopTask, public Signaller {
public:
    SignalTask(TQP(tqp) const char* ident, ms_t cadence);
    void signal(int, ms_t) override;
};
SignalTask::SignalTask(TQP(tqp) const char* ident, ms_t cadence)
    : TestLoopTask(TQPX(tqp) ident, cadence, -1),
      Signaller(ident)
{
    // empty
}
void SignalTask::signal(int signal, ms_t t)
{
    Signaller::signal(signal, t);
}

void test_signal_task(void)
{
    reset_all();

    // start the tasks with no end-times
    SignalTask s1(TQPP "s1", 5);
    SignalTask s2(TQPP "s2", 4);
    // ordinary TestLoopTask, checking they can co-exist
    TestLoopTask lt(TQPP "lt", 7);
    s1.start();
    s2.start();
    lt.start();

    TerminateTask x;
    x.start(18);

    for (ms_t t = 0; !Q.is_empty(); t++) {
        if (t == 6 || t == 11) {
            Q.signal(1, t);
        }

        Q.run_ready(t);
    }

    assert_equal("s1@0,s2@1,lt@2," // first time -- collisions
                 "s2@4,s1@5,"      // signal tasks run
                 "Ss1@6,Ss2@6,"    // first signal
                 "lt@7,s2@8,s1@10,"
                 "Ss1@11,Ss2@11,"  // second signal
                 "s2@12,lt@14,s1@15,s2@16",
                 log_buffer);
}

void test_signal_requeue(void)
{
    class SignalAddTask : public TestTask, public Signaller {
        // This task enqueues a second task (provided to it),
        // when it itself receives a signal
    private:
        ms_t t2;
        Task* T2;
    public:
        SignalAddTask(TQP(tqp) const char* ident, Task* T2, ms_t t2)
            : TestTask(TQPX(tqp) ident),
              Signaller(ident),
              t2(t2),
              T2(T2) {}
        void signal(int signal, ms_t t) override {
            Signaller::signal(signal, t);
            T2->start(t2);
        };
    };

    // If a task receives a signal, it may respond to this by
    // enqueueing further tasks.  These new tasks should NOT receive
    // that signal.
    SignalTask t1 = SignalTask(TQPP "t1", 0);
    SignalTask t2 = SignalTask(TQPP "t2", 0);
    SignalAddTask s1 = SignalAddTask(TQPP "s1", &t1, 10);
    SignalAddTask s2 = SignalAddTask(TQPP "s2", &t2, 0);

    reset_all();

    s1.start(5);
    s2.start(8);

    for (ms_t t=0; !Q.is_empty(); t++) {
        if (t == 3) {
            Q.signal(1, t);
        }

        Q.run_ready(t);
    }

    // The library doesn't specify what order s1 and s2 receive the
    // signal, but it is consistent
    assert_equal("Ss2@3,Ss1@3," // only -- ie, no St1 or St2
                 "t2@3,s1@5,s2@8,t1@13",
                 log_buffer);
}

void test_never_task(void)
{
    reset_all();

    TestTask t1 = TestTask(TQPP "t1");
    SignalTask t2 = SignalTask(TQPP "t2", 0);
    TerminateTask tx;

    t1.start(5);
    t2.start(TIME_NEVER);       // should never run, but can receive signals
    tx.start(10);

    for (ms_t t=0; !Q.is_empty(); t++) {
        if (t == 7) {
            Q.signal(1, t);
        }
        Q.run_ready(t);
    }

    assert_log("t1@5,St2@7");
}

// This is related to test_reference_times, but is here because we want to be able to use the SignalTask task.
void test_reference_times_after(void)
{
    reset_all();

    TestTask t1 = TestTask(TQPP "t1");
    TestTask t2 = TestTask(TQPP "t2");
    TestTask t3 = TestTask(TQPP "t3");

    t1.start();
    t2.start(3);

    SignalTask s1(TQPP "s1", 999);
    s1.start();                 // will run only once

    TerminateTask x;
    x.start(30);

    for (ms_t t=0; !Q.is_empty(); t++) {
        switch (t) {
          case 5:
            // Task t3 should be run at time 3+7=10, not 5+7=12.
            // This is because the 'reference time' is set by the last non-NULL pop.
            Q.push_after(&t3, 7);
            break;

          case 12:
            Q.signal(1, t);
            break;

          case 14:
            // Task t3 should next be run at time 12+3,
            // because the 'reference time' is set by the signal.
            Q.push_after(&t3, 3);
            break;
        }

        Q.run_ready(t);
    }

    assert_log("t1@0,s1@1,t2@3,t3@10,Ss1@12,t3@15");
}

#if PQRST_TASK_PRIORITIES
void test_priority_task(void)
{
    reset_all();

    TestTask n1 = TestTask(TQPP "n1");
    TestTask n2 = TestTask(TQPP "n2");
    TestTask pt = TestTask(TQPP "pt");
    TestTask pt2 = TestTask(TQPP "pt2");

    assert_equal(1, n1.get_priority()); // tasks are by default priority tasks of duration 1
    assert_equal(1, n1.get_duration());

    n1.set_priority(false);
    n1.set_duration(4);
    n2.set_priority(false);
    n2.set_duration(4);
    pt.set_priority(false);
    pt.set_duration(4);
    pt.set_priority(true);
    pt2.set_duration(6);        // longer duration than pt
    pt2.set_priority(true);

    reset_all();
    n1.start(3);
    pt.start(6);                // priority task added second, but should run first
    run_queue();
    assert_log("pt@6,n1@7");

    reset_all();
    n1.start(3);
    pt.start(7);                // late enough to be clear
    run_queue();
    assert_log("n1@3,pt@7");

    reset_all();
    n1.start(3);
    n2.start(5);
    pt.start(7);                // clear of first, conflicts with second
    run_queue();
    assert_log("n1@3,pt@7,n2@8");

    reset_all();
    n1.start(3);
    n2.start(5);
    pt.start(9);                // clear of both
    run_queue();
    assert_log("n1@3,n2@5,pt@9");

    reset_all();
    pt.start(5);                // priority added first
    n1.start(3);
    n2.start(4);
    run_queue();
    assert_log("pt@5,n1@6,n2@7");

    reset_all();
    pt.start(5);
    n1.start(1);                // clear
    n2.start(2);                // not clear
    run_queue();
    assert_log("n1@1,pt@5,n2@6");

    reset_all();
    // Note: in this case, the pt and pt2 tasks would actually be
    // delayed in the real case, since the two low-priority tasks
    // would run for a total of 8ms; but we don't notice this here,
    // since the algorithm only compares neighbouring tasks, and
    // doesn't add up durations.
    n1.start(1);
    n2.start(1);                // scheduled at same time
    pt.start(5);
    pt2.start(5);               // scheduled at same time
    run_queue();
    assert_log("n1@1,n2@2,pt@5,pt2@6");

    reset_all();
    n2.start(1);
    n1.start(1);
    pt2.start(5);               // longer task scheduled first
    pt.start(5);
    run_queue();
    assert_log("n2@1,n1@2,pt@5,pt2@6");
}
#endif

#if PQRST_TASK_IDENT
void traversal_callback(ms_t t, const char* UNUSED(id), const Task* T)
#else
void traversal_callback(ms_t t, const Task* T)
#endif
{
    if (T->user_task_p()) {
        char buf[16];
        snprintf(buf, 16, "cb-" MS_FMT, t);
        buf_append_ident(buf, 0);
    }
}

void test_traversal(void)
{
    reset_all();

    TestTask t1 = TestTask(TQPP "t1");
    TestTask t2 = TestTask(TQPP "t2");
    TestTask t3 = TestTask(TQPP "t3");

    t1.start();
    t2.start(5);
    t3.start(99);

    Q.traverse_queue(traversal_callback);

    assert_log("cb-0@0,cb-5@0,cb-99@0");
}

void test_multiple_insertion(void)
{
    TestTask t1 = TestTask(TQPP "t1");
    TestTask t2 = TestTask(TQPP "t2");

    reset_all();
    t1.start(1);
    t2.start(5);
    t2.start(10);               // inserted twice, second later (at end), first should be ignored
    run_queue();
    assert_log("t1@1,t2@10");

    reset_all();
    t1.start(1);
    t2.start(10);
    t2.start(5);                // second is earlier, first should be ignored
    run_queue();
    assert_log("t1@1,t2@5");

    reset_all();
    t1.start(5);
    t2.start(10);
    t2.start(1);                // second inserted at front of queue
    run_queue();
    assert_log("t2@1,t1@5");

    reset_all();
    t1.start(5);
    t2.start(1);
    t2.start(10);               // second replaces item at front of queue
    run_queue();
    assert_log("t1@5,t2@10");
}

// Tests of time overflowing.
//
// The ms_t type (32-bit unsigned integer) will overflow at time
// 0xffffffff (=4294967295), half of which is 0x7fffffff
// (=2147483647).  See the discussion in pqrst.h.

void test_march_of_time(void)
{
    reset_all();

    TestLoopTask tt = TestLoopTask(TQPP "t", 0x40000000, 7);
#if PQRST_TASK_IDENT
    tt.set_task_ident("tt");
#endif
    tt.start();

    run_queue_with_repeats(0, 0x10000000);

    assert_log("t@0,t@0x40000000,t@0x80000000,t@0xc0000000,t@0,t@0x40000000,t@0x80000000");
}

void test_tiptoe_of_time(void)
{
    // as with test_march_of_time, but tiptoe over the relevant boundaries
    reset_all();

    TestLoopTask tt = TestLoopTask(TQPP "t", 1, 5);
    tt.start(HALF_TIME - 2);    // fffffffe

    run_queue_with_repeats(HALF_TIME - 8, 1);

    assert_log("t@0x7ffffffe,t@0x7fffffff,t@0x80000000,t@0x80000001,t@0x80000002");

    // now the same, but stepping over the 0xffffffff -> 0 overflow
    reset_all();

    tt = TestLoopTask(TQPP "t", 1, 5);
    tt.start(FULL_TIME - 1);

    // step over the 0x80000000 call to ModularTimeMinder,
    // in order to make sure that it is called in the expected
    // sequence
    // (see also test_time_wrapping_late, which tests that this
    // caution is unnecessary)
    for (ms_t t1 = 0x7fffffff; t1<0x80000001; t1++) {
        Q.run_ready(t1);
    }

    // now jump to the end of the high-time
    run_queue_with_repeats(FULL_TIME - 8, 1);

    assert_log("t@0xfffffffe,t@0xffffffff,t@0,t@1,t@2");
}

void test_time_wrapping_late(void)
{
    // Test the case where the first non-maintenance task scheduled is
    // in high time, above 0x80000000, and we start timing after that
    // half-way point.  This might happen in practice if the queue is
    // reinitialised at a time when millis() or micros() is in high-time.
    //
    // The ModularTimeMinder object shouldn't become confused about this.
    reset_all();

    TestLoopTask tt = TestLoopTask(TQPP "t", 0x20000000, 5);
    tt.start(0xb0000000);       // after the half-way point, 0x80000000

    run_queue_with_repeats(0x90000000, 0x10000000);

    assert_log("t@0xb0000000,t@0xd0000000,t@0xf0000000,t@0x10000000,t@0x30000000");
}

void test_next_due(void)
{
    reset_all();

    assert_equal(FULL_TIME, Q.next_due_after()); // empty queue

    TestTask t1 = TestTask(TQPP "t1");
    TestTask t2 = TestTask(TQPP "t2");

    t1.run_at(1000);
    assert_equal(1000, Q.next_due_after());
    Q.run_ready(1000);          // reftime now 1000

    t2.run_at(500);                      // previous to reftime
    assert_equal(0, Q.next_due_after()); // not negative
    Q.run_ready(1000);                   // remove t2 from queue

    t1.run_at(1000);
    t2.run_at(HALF_TIME + 1000); // in upper half of time range

    Q.run_ready(1000);          // remove t1; reftime now 1000
    assert_equal(HALF_TIME, Q.next_due_after());
    run_queue(HALF_TIME-500, 500);
    // remove (modtime minder and) t2; reftime now t2 ready-time: HALF_TIME + 1000

    t1.run_at(1000);            // after wraparound
    assert_equal(HALF_TIME, Q.next_due_after());

    //printf("buffer: %s\n", log_buffer);
}

void test_ready_after(void)
{
    class CheckAfter : public TestTask {
    private:
        TestTask& other_;
        ms_t expect_after_;
        int from_line_;
    public:
        CheckAfter(TestTask& tt, ms_t expect_after, int l)
            : TestTask(TQPP "ca"), other_(tt), expect_after_(expect_after), from_line_(l) {};
        void run(ms_t UNUSED(t)) {
            assert_equal_(__FILE__, from_line_, other_.ready_after(), expect_after_);
        };
    };

    reset_all();

    TestTask t1(TQPP "t1");

    // unqueued task
    assert_equal(t1.ready_after(), TIME_NEVER);

    // low time
    t1.run_at(100);
    CheckAfter(t1, 50, __LINE__).run_at(50);
    run_queue();

    // time straddling half-time
    t1.run_at(HALF_TIME + 50);
    CheckAfter(t1, 100, __LINE__).run_at(HALF_TIME - 50);
    run_queue(HALF_TIME - 60);
    // queue is now in high time

    // high time for both tasks
    t1.run_at(HALF_TIME + 100);
    CheckAfter(t1, 50, __LINE__).run_at(HALF_TIME + 50);
    run_queue(HALF_TIME + 10);

    // straddling roll-over
    t1.run_at(50);
    CheckAfter(t1, 100, __LINE__).run_at(0 - 50);
    run_queue(FULL_TIME - 100);
}

void test_lateness(void)
{
    reset_all();

#if PQRST_SELF_QUEUE
    // the particular form of this test, with (Q.),
    // doesn't work in FunctionTask
    assert_pass();
#else

#define LATE_AT(run_time, expd)                 \
    FunctionTask([](ms_t t) {                   \
            assert_equal(expd, Q.lateness(t));  \
        }).run_at(run_time)

    LATE_AT(5, 0);
    // the next task is scheduled at the same time, so will be run late
    LATE_AT(5, 1);
    run_queue();

    LATE_AT(HALF_TIME - 2, 0);
    LATE_AT(HALF_TIME - 2, 1);
    LATE_AT(HALF_TIME + 2, 0);
    LATE_AT(HALF_TIME + 2, 1);
    run_queue(HALF_TIME - 4, 1);

    LATE_AT(FULL_TIME, 0);
    LATE_AT(FULL_TIME, 1);      // will run at time 0
    run_queue(FULL_TIME - 4, 1);

#undef LATE_AT
#endif
}

void test_reinitialisation(void)
{
    class ShowLatenessTask : public TestLoopTask {
    public:
        ShowLatenessTask(TQP(tqp) const char* ident) : TestLoopTask(TQPX(tqp) ident, 5, 2) {};
        void run(ms_t t) {
            const size_t buflen = 30;
            char buf[buflen];
            snprintf(buf, buflen, "%s-%d", ident(), Q.lateness(t));
            buf_append_ident(buf, t);
        };
    };

    reset_all();

    // without this (re)initialisation of the queue reftime,
    // the tasks below would (a) all be around 100 time units late,
    // and (b) all in successive timeslots.
    Q.initialise(100);

    ShowLatenessTask t1 = ShowLatenessTask(TQPP "t1");
    ShowLatenessTask t2 = ShowLatenessTask(TQPP "t2");

    t1.start();
    t2.start(20);

    run_queue(100);

    assert_log("t1-0@100,t1-0@105,t2-0@120,t2-0@125");
}

int main(int argc, char** argv)
{
    //printf("Testing %s\n", Task::pqrst_version);

    Testcase funcs[] = {
        { "tasks in order", test_in_order, },
        { "function task", test_function_task, },
        { "function task2", test_function_task2, },
        { "tasks out of order", test_out_of_order, },
        { "weird calls", test_weird_calls, },
        { "simultaneous tasks", test_ordering, },
        { "removing tasks", test_removals, },
        { "removing and uncancelling tasks", test_uncancel, },
        { "removing tasks (all)", test_remove_all, },
        { "repeating tasks", test_looping, },
        { "repeating indefinitely", test_indefinite_loop, },
        { "repeating indefinitely (bis)", test_indefinite_loop_pair, },
        { "repeating curtailed", test_adjusted_loop, },
        { "repeating non-loop", test_nonloop_loop, },
        { "repeating zero-cadence", test_zerocadence_loop, },
        { "repeating stuttering", test_loop_stutter, },
        { "never-run task", test_never_task, },
        { "reference times (push_at)", test_reference_times, },
        { "reference times (push_after)", test_reference_times_after, },
        { "reference times (slippy)", test_reference_times_slippy, },
        { "self-cancel/helper", test_self_cancel_task, },
        { "self-cancel/self", test_self_cancel_task2, },
        { "signal empty queue", test_signal_when_empty, },
        { "signal simple", test_signal_task, },
        { "signal requeueing", test_signal_requeue, },
#if PQRST_TASK_PRIORITIES
        { "priority tasks", test_priority_task, },
#endif
        { "traversal", test_traversal, },
        { "multiple insertion", test_multiple_insertion, },
        { "time wrapping fast", test_march_of_time, },
        { "time wrapping slow", test_tiptoe_of_time, },
        { "time wrapping late", test_time_wrapping_late, },
        { "next due", test_next_due, },
        { "lateness", test_lateness, },
        { "reinitialisation", test_reinitialisation, },
        { "ready_after()", test_ready_after, },
    };
    int nfuncs = sizeof(funcs)/sizeof(funcs[0]);

    return run_test_cases(nullptr, funcs, nfuncs, argc, argv);
}
